"""ORM models for legacy user and session tables."""

from typing import Any, NewType, List

from sqlalchemy import BigInteger, Column, DateTime, Enum, \
    ForeignKey, Index, Integer, SmallInteger, String, Text, text
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

from arxiv import taxonomy
from .. import domain

Base = declarative_base()


class DBSession(Base):  # type: ignore
    """
    Legacy arXiv session table.

    +----------------+-----------------+------+-------+---------+
    | Field          | Type            | Null | Key   | Default |
    +----------------+-----------------+------+-------+---------+
    | session_id     | int(4) unsigned | NO   | PRI   | NULL    |
    | user_id        | int(4) unsigned | NO   | MUL   | 0       |
    | last_reissue   | int(11)         | NO   |       | 0       |
    | start_time     | int(11)         | NO   | MUL   | 0       |
    | end_time       | int(11)         | NO   | MUL   | 0       |
    +--------------+-------------------+------+-------+---------+
    """

    __tablename__ = 'tapir_sessions'

    session_id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(ForeignKey('tapir_users.user_id'), nullable=False,
                     index=True, server_default=text("'0'"))
    last_reissue = Column(Integer, nullable=False, server_default=text("'0'"))
    start_time = Column(Integer, nullable=False, index=True,
                        server_default=text("'0'"))
    end_time = Column(Integer, nullable=False, index=True,
                      server_default=text("'0'"))

    user = relationship('DBUser')


class DBSessionsAudit(Base):  # type: ignore
    """Legacy arXiv session audit table. Notably has a tracking cookie."""

    __tablename__ = 'tapir_sessions_audit'

    session_id = Column(
        ForeignKey('tapir_sessions.session_id'), primary_key=True,
        autoincrement="false", server_default=text("'0'")
    )
    ip_addr = Column(String(16), nullable=False, index=True,
                     server_default=text("''"))
    remote_host = Column(String(255), nullable=False,
                         server_default=text("''"))
    tracking_cookie = Column(String(255), nullable=False, index=True,
                             server_default=text("''"))

    session = relationship('DBSession')


class DBUser(Base):  # type: ignore
    """Legacy user data table."""

    __tablename__ = 'tapir_users'

    user_id = Column(Integer, primary_key=True)
    first_name = Column(String(50), index=True)
    last_name = Column(String(50), index=True)
    suffix_name = Column(String(50))
    share_first_name = Column(Integer, nullable=False, server_default=text("'1'"))
    share_last_name = Column(Integer, nullable=False, server_default=text("'1'"))
    email = Column(String(255), nullable=False, unique=True, server_default=text("''"))
    share_email = Column(Integer, nullable=False, server_default=text("'8'"))
    email_bouncing = Column(Integer, nullable=False, server_default=text("'0'"))
    policy_class = Column(
        ForeignKey('tapir_policy_classes.class_id'),
        nullable=False, index=True, server_default=text("'0'")
    )
    joined_date = Column(Integer, nullable=False, index=True, server_default=text("'0'"))
    joined_ip_num = Column(String(16), index=True)
    joined_remote_host = Column(String(255), nullable=False, server_default=text("''"))
    flag_internal = Column(Integer, nullable=False, index=True, server_default=text("'0'"))
    flag_edit_users = Column(Integer, nullable=False, index=True, server_default=text("'0'"))
    flag_edit_system = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_email_verified = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_approved = Column(Integer, nullable=False, index=True, server_default=text("'1'"))
    flag_deleted = Column(Integer, nullable=False, index=True, server_default=text("'0'"))
    flag_banned = Column(Integer, nullable=False, index=True, server_default=text("'0'"))
    flag_wants_email = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_html_email = Column(Integer, nullable=False, server_default=text("'0'"))
    tracking_cookie = Column(String(255), nullable=False, index=True, server_default=text("''"))
    flag_allow_tex_produced = Column(Integer, nullable=False, server_default=text("'0'"))


class DBPolicyClass(Base):  # type: ignore
    """Legacy authorization table."""

    __tablename__ = 'tapir_policy_classes'

    ADMIN = 1
    PUBLIC_USER = 2
    LEGACY_USER = 3
    POLICY_CLASSES = [
        {"name": "Administrator", "class_id": ADMIN, "description": ""},
        {"name": "Public user", "class_id": PUBLIC_USER, "description": ""},
        {"name": "Legacy user", "class_id": LEGACY_USER, "description": ""}
    ]

    class_id = Column(SmallInteger, primary_key=True)
    name = Column(String(64), nullable=False, server_default=text("''"))
    description = Column(Text, nullable=False)
    password_storage = Column(Integer, nullable=False,
                              server_default=text("'0'"))
    recovery_policy = Column(Integer, nullable=False,
                             server_default=text("'0'"))
    permanent_login = Column(Integer, nullable=False,
                             server_default=text("'0'"))


class DBUserPassword(Base):  # type: ignore
    """Legacy password table."""

    __tablename__ = 'tapir_users_password'

    user_id = Column(ForeignKey('tapir_users.user_id'), nullable=False,
                     server_default=text("'0'"), primary_key=True)
    password_storage = Column(Integer, nullable=False, index=True,
                              server_default=text("'0'"))
    password_enc = Column(String(16), nullable=False)

    user = relationship('DBUser')


class DBPermanentToken(Base):  # type: ignore
    """
    Bearer token for user authentication.

    +-------------+-----------------+------+-----+---------+-------+
    | Field       | Type            | Null | Key | Default | Extra |
    +-------------+-----------------+------+-----+---------+-------+
    | user_id     | int(4) unsigned | NO   | PRI | 0       |       |
    | secret      | varchar(32)     | NO   | PRI |         |       |
    | valid       | int(1)          | NO   |     | 1       |       |
    | issued_when | int(4) unsigned | NO   |     | 0       |       |
    | issued_to   | varchar(16)     | NO   |     |         |       |
    | remote_host | varchar(255)    | NO   |     |         |       |
    | session_id  | int(4) unsigned | NO   | MUL | 0       |       |
    +-------------+-----------------+------+-----+---------+-------+
    """

    __tablename__ = 'tapir_permanent_tokens'

    user_id = Column(Integer, primary_key=True)
    secret = Column(String(32), primary_key=True)
    """Token."""
    valid = Column(Integer, nullable=False, server_default=text("'1'"))
    issued_when = Column(Integer, nullable=False, server_default=text("'0'"))
    """Epoch time."""
    issued_to = Column(String(16), nullable=False)
    """IP address of client."""
    remote_host = Column(String(255), nullable=False)
    session_id = Column(Integer, nullable=False, server_default=text("'0'"))


class DBUserNickname(Base):  # type: ignore
    """
    Users' usernames (because why not have a separate table).

    +--------------+------------------+------+-----+---------+----------------+
    | Field        | Type             | Null | Key | Default | Extra          |
    +--------------+------------------+------+-----+---------+----------------+
    | nick_id      | int(10) unsigned | NO   | PRI | NULL    | autoincrement  |
    | nickname     | varchar(20)      | NO   | UNI |         |                |
    | user_id      | int(4) unsigned  | NO   | MUL | 0       |                |
    | user_seq     | int(1) unsigned  | NO   |     | 0       |                |
    | flag_valid   | int(1) unsigned  | NO   | MUL | 0       |                |
    | role         | int(10) unsigned | NO   | MUL | 0       |                |
    | policy       | int(10) unsigned | NO   | MUL | 0       |                |
    | flag_primary | int(1) unsigned  | NO   |     | 0       |                |
    +--------------+------------------+------+-----+---------+----------------+
    """

    __tablename__ = 'tapir_nicknames'

    nick_id = Column(Integer, primary_key=True)
    nickname = Column(String(20), nullable=False, unique=True, index=True)
    user_id = Column(ForeignKey('tapir_users.user_id'), nullable=False,
                     server_default=text("'0'"))
    user_seq = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_valid = Column(Integer, nullable=False, server_default=text("'0'"))
    role = Column(Integer, nullable=False, server_default=text("'0'"))
    policy = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_primary = Column(Integer, nullable=False, server_default=text("'0'"))

    user = relationship('DBUser')


# TODO: update based on recent schema changes.
class DBProfile(Base):   # type: ignore
    """Legacy user profiles."""

    __tablename__ = 'arxiv_demographics'

    TYPE_CHOICES = [
        (1, 'Staff'),
        (2, "Professor"),
        (3, "Post Doc"),
        (4, "Grad Student"),
        (5, "Other")
    ]
    """Legacy ranks in arXiv user profiles."""

    user_id = Column(ForeignKey('tapir_users.user_id'), nullable=False,
                     server_default=text("'0'"), primary_key=True)
    country = Column(String(2), nullable=False)
    affiliation = Column(String(255), nullable=False)
    url = Column(String(255), nullable=False)
    rank = Column('type', SmallInteger, nullable=True, server_default=text("'null'"))
    archive = Column(String(16), nullable=True, server_default=text("'null'"))
    subject_class = Column(String(16), nullable=True, server_default=text("'null'"))
    original_subject_classes = Column(String(255), nullable=False)
    flag_group_physics = Column(Integer, nullable=True, server_default=text("'null'"))
    flag_group_math = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_cs = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_nlin = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_q_bio = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_q_fin = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_stat = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_econ = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_group_eess = Column(Integer, nullable=False, server_default=text("'0'"))
    user = relationship('DBUser')

    GROUP_FLAGS = [
        ('grp_physics', 'flag_group_physics'),
        ('grp_math', 'flag_group_math'),
        ('grp_cs', 'flag_group_cs'),
        ('grp_q-bio', 'flag_group_q_bio'),
        ('grp_q-fin', 'flag_group_q_fin'),
        ('grp_q-stat', 'flag_group_stat'),
        ('grp_q-econ', 'flag_group_econ'),
        ('grp_q-eess', 'flag_group_eess'),
    ]

    @property
    def groups(self) -> List[str]:
        """Active groups for this user profile."""
        return [group for group, column in self.GROUP_FLAGS
                if getattr(self, column) == 1]

    def to_domain(self) -> domain.UserProfile:
        """Generate a domain representation from this database instance."""
        return domain.UserProfile(  # type: ignore
            affiliation=self.affiliation,
            country=self.country,
            rank=self.rank,
            submission_groups=self.groups,
            default_category=domain.Category(self.archive, self.subject_class),
            homepage_url=self.url,
        )


class DBEndorsement(Base):  # type: ignore
    """
    Category endorsements for arXiv users.

    +----------------+-----------------------------+------+-----+---------+
    | Field          | Type                        | Null | Key | Default |
    +----------------+-----------------------------+------+-----+---------+
    | endorsement_id | int(10) unsigned            | NO   | PRI | NULL    |
    | endorser_id    | int(10) unsigned            | YES  | MUL | NULL    |
    | endorsee_id    | int(10) unsigned            | NO   | MUL | 0       |
    | archive        | varchar(16)                 | NO   | MUL |         |
    | subject_class  | varchar(16)                 | NO   |     |         |
    | flag_valid     | int(1) unsigned             | NO   |     | 0       |
    | type           | enum('user','admin','auto') | YES  |     | NULL    |
    | point_value    | int(1) unsigned             | NO   |     | 0       |
    | issued_when    | int(10) unsigned            | NO   |     | 0       |
    | request_id     | int(10) unsigned            | YES  | MUL | NULL    |
    +----------------+-----------------------------+------+-----+---------+
    """

    __tablename__ = 'arXiv_endorsements'

    endorsement_id = Column(Integer, primary_key=True, nullable=False)
    endorser_id = Column(Integer, nullable=True, server_default=None)
    endorsee_id = Column(ForeignKey('tapir_users.user_id'), nullable=False,
                         server_default=text("'0'"))
    archive = Column(String(16), nullable=False)
    subject_class = Column(String(16), nullable=False)
    flag_valid = Column(Integer, nullable=False, server_default=text("'0'"))
    endorsement_type = Column('type', Enum('user', 'admin', 'auto'),
                              nullable=True, server_default=None)
    point_value = Column(Integer, nullable=False, server_default=text("'0'"))
    issued_when = Column(Integer, nullable=False, server_default=text("'0'"))
    request_id = Column(Integer, nullable=True, server_default=None)

    endorsee = relationship('DBUser')


class DBEndorsementDomain(Base):  # type: ignore
    """
    Encodes some policies about endorsement.

    +--------------------+----------------------+------+-----+---------+
    | Field              | Type                 | Null | Key | Default |
    +--------------------+----------------------+------+-----+---------+
    | endorsement_domain | varchar(32)          | NO   | PRI |         |
    | endorse_all        | enum('y','n')        | NO   |     | n       |
    | mods_endorse_all   | enum('y','n')        | NO   |     | n       |
    | endorse_email      | enum('y','n')        | NO   |     | y       |
    | papers_to_endorse  | smallint(5) unsigned | NO   |     | 4       |
    +--------------------+----------------------+------+-----+---------+
    """

    __tablename__ = 'arXiv_endorsement_domains'

    endorsement_domain = Column(String(32), primary_key=True)
    endorse_all = Column(Enum('y', 'n'), server_default='n')
    mods_endorse_all = Column(Enum('y', 'n'), server_default='n')
    endorse_email = Column(Enum('y', 'n'), server_default='n')
    papers_to_endorse = Column(Integer, nullable=False,
                               server_default=text("'4'"))


class DBCategory(Base):  # type: ignore
    """
    Metadata about arXiv categories.

    +--------------------+----------------------+------+-----+---------+
    | Field              | Type                 | Null | Key | Default |
    +--------------------+----------------------+------+-----+---------+
    | archive            | varchar(16)          | NO   | PRI |         |
    | subject_class      | varchar(16)          | NO   | PRI |         |
    | definitive         | int(1)               | NO   |     | 0       |
    | active             | int(1)               | NO   |     | 0       |
    | category_name      | varchar(255)         | YES  |     | NULL    |
    | endorse_all        | enum('y','n','d')    | NO   |     | d       |
    | endorse_email      | enum('y','n','d')    | NO   |     | d       |
    | papers_to_endorse  | smallint(5) unsigned | NO   |     | 0       |
    | endorsement_domain | varchar(32)          | YES  | MUL | NULL    |
    +--------------------+----------------------+------+-----+---------+
    """

    __tablename__ = 'arXiv_categories'

    archive = Column(String(16), primary_key=True)
    subject_class = Column(String(16), primary_key=True)
    definitive = Column(Integer, nullable=False, server_default=text("'0'"))
    active = Column(Integer, nullable=False, server_default=text("'0'"))
    endorsement_domain = Column(String(32), nullable=True)


class DBPaperOwners(Base):  # type: ignore
    """
    Relates arXiv users to their owned papers.

    +-----------------+-----------------------+------+-----+---------+-------+
    | Field           | Type                  | Null | Key | Default | Extra |
    +-----------------+-----------------------+------+-----+---------+-------+
    | document_id     | mediumint(8) unsigned | NO   | PRI | 0       |       |
    | user_id         | int(10) unsigned      | NO   | PRI | 0       |       |
    | date            | int(10) unsigned      | NO   |     | 0       |       |
    | valid           | int(1) unsigned       | NO   |     | 0       |       |
    | flag_author     | int(1) unsigned       | NO   |     | 0       |       |
    +-----------------+-----------------------+------+-----+---------+-------+
    """

    __tablename__ = 'arXiv_paper_owners'

    document_id = Column(Integer, primary_key=True)
    user_id = Column(Integer, primary_key=True)
    date = Column(Integer, nullable=False, server_default=text("'0'"))
    flag_author = Column(Integer, nullable=False, server_default=text("'0'"))
    valid = Column(Integer, nullable=False, server_default=text("'0'"))


class DBDocuments(Base):  # type: ignore
    """
    Represents an arXiv paper.

    +-----------------------+-----------------------+------+-----+---------+
    | Field                 | Type                  | Null | Key | Default |
    +-----------------------+-----------------------+------+-----+---------+
    | document_id           | mediumint(8) unsigned | NO   | PRI | NULL    |
    | paper_id              | varchar(20)           | NO   | UNI |         |
    | title                 | varchar(255)          | NO   | MUL |         |
    | authors               | text                  | YES  |     | NULL    |
    | submitter_email       | varchar(64)           | NO   | MUL |         |
    | submitter_id          | int(10) unsigned      | YES  | MUL | NULL    |
    | dated                 | int(10) unsigned      | NO   | MUL | 0       |
    | primary_subject_class | varchar(16)           | YES  |     | NULL    |
    | created               | datetime              | YES  |     | NULL    |
    +-----------------------+-----------------------+------+-----+---------+
    """

    __tablename__ = 'arXiv_documents'

    document_id = Column(Integer, primary_key=True, autoincrement=True)
    paper_id = Column(String(20), nullable=False, unique=True)
    dated = Column(Integer, nullable=False, server_default=text("'0'"))


class DBDocumentInCategory(Base):  # type: ignore
    """
    M2M intermediate table for documents and their categories.

    +---------------+-----------------------+------+-----+---------+-------+
    | Field         | Type                  | Null | Key | Default | Extra |
    +---------------+-----------------------+------+-----+---------+-------+
    | document_id   | mediumint(8) unsigned | NO   | PRI | 0       |       |
    | archive       | varchar(16)           | NO   | PRI |         |       |
    | subject_class | varchar(16)           | NO   | PRI |         |       |
    | is_primary    | tinyint(1)            | NO   |     | 0       |       |
    +---------------+-----------------------+------+-----+---------+-------+
    """

    __tablename__ = 'arXiv_in_category'

    document_id = Column(Integer, primary_key=True)
    archive = Column(String(16), primary_key=True)
    subject_class = Column(String(16), primary_key=True)
    is_primary = Column(Integer, nullable=False, server_default=text("'0'"))


class DBEmailWhitelist(Base):  # type: ignore
    """
    Patterns for identifying academic addresses.

    pattern | varchar(64) | YES  |     | NULL    |
    """

    __tablename__ = 'arXiv_white_email'

    pattern = Column(String(64), primary_key=True)


class DBEmailBlacklist(Base):  # type: ignore
    """
    Patterns for identifying non-academic addresses.

    pattern | varchar(64) | YES  |     | NULL    |
    """

    __tablename__ = 'arXiv_black_email'

    pattern = Column(String(64), primary_key=True)



"""Provides API for legacy user sessions."""

import ipaddress
import json
from datetime import datetime, timedelta
from pytz import timezone
import hashlib
from base64 import b64encode, b64decode

from typing import Optional, Generator, Tuple, List

from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

from arxiv.base import logging

from .. import domain
from . import cookies, util

from .models import Base, DBSession, DBSessionsAudit, DBUser, DBEndorsement, \
    DBUserNickname
from .exceptions import UnknownSession, SessionCreationFailed, \
    SessionDeletionFailed, SessionExpired, InvalidCookie
from .endorsements import get_endorsements

logger = logging.getLogger(__name__)
EASTERN = timezone('US/Eastern')


def _load(session_id: str) -> DBSession:
    """Get DBSession from session id."""
    with util.transaction() as session:
        db_session: DBSession = session.query(DBSession) \
            .filter(DBSession.session_id == session_id) \
            .first()
    if not db_session:
        logger.error(f'No session found with id {session_id}')
        raise UnknownSession('No such session')
    return db_session


def load(cookie: str) -> domain.Session:
    """
    Given a session cookie (from request), load the logged-in user.

    Parameters
    ----------
    cookie : str
        Legacy cookie value passed with the request.

    Returns
    -------
    :class:`.domain.Session`

    Raises
    ------
    :class:`.legacy.exceptions.SessionExpired`
    :class:`.legacy.exceptions.UnknownSession`

    """
    session_id, user_id, ip, issued_at, expires_at, _ = cookies.unpack(cookie)
    logger.debug('Load session %s for user %s at %s',
                 session_id, user_id, ip)

    if expires_at <= datetime.now(tz=EASTERN):
        raise SessionExpired(f'Session {session_id} has expired')

    with util.transaction() as session:
        data: Tuple[DBUser, DBSession, DBUserNickname] = (
            session.query(DBUser, DBSession, DBUserNickname)
            .filter(DBUser.user_id == user_id)
            .filter(DBUserNickname.user_id == user_id)
            .filter(DBSession.user_id == user_id)
            .filter(DBSession.session_id == session_id)
            .first()
        )

        if not data:
            raise UnknownSession('No such user or session')

        db_user, db_session, db_nick = data

        # Verify that the session is not expired.
        if db_session.end_time != 0 and db_session.end_time < util.now():
            logger.info('Session has expired: %s', session_id)
            raise SessionExpired(f'Session {session_id} has expired')

        user = domain.User(
            user_id=str(user_id),
            username=db_nick.nickname,
            email=db_user.email,
            name=domain.UserFullName(
                forename=db_user.first_name,
                surname=db_user.last_name,
                suffix=db_user.suffix_name
            )
        )

        # We should get one row per endorsement.
        authorizations = domain.Authorizations(
            classic=util.compute_capabilities(db_user),
            endorsements=get_endorsements(user),
            scopes=util.get_scopes(db_user)
        )

    user_session = domain.Session(str(db_session.session_id),
                                  start_time=issued_at, end_time=expires_at,
                                  user=user, authorizations=authorizations)
    logger.debug('loaded session %s', user_session.session_id)
    return user_session


def create(user: domain.User, authorizations: domain.Authorizations,
           ip: str, remote_host: str, tracking_cookie: str = '') \
        -> Tuple[domain.Session, str]:
    """
    Create a new legacy session for an authenticated user.

    Parameters
    ----------
    user : :class:`.User`
    ip : str
        Client IP address.
    remote_host : str
        Client hostname.
    tracking_cookie : str
        Tracking cookie payload from client request.

    Returns
    -------
    :class:`.Session`

    """
    logger.debug('create session for user %s', user.user_id)
    start = datetime.now(tz=EASTERN)
    end = start + timedelta(seconds=util.get_session_duration())
    try:
        with util.transaction() as session:
            tapir_session = DBSession(
                user_id=user.user_id,
                last_reissue=0,   # TODO: do we need this?
                start_time=util.epoch(start),
                end_time=0
            )
            tapir_sessions_audit = DBSessionsAudit(
                session=tapir_session,
                ip_addr=ip,
                remote_host=remote_host,
                tracking_cookie=tracking_cookie
            )
            session.add(tapir_sessions_audit)
            session.commit()
    except Exception as e:  # TODO: be more specific.
        raise SessionCreationFailed(f'Failed to create: {e}') from e

    user_id = '' if user.user_id is None else user.user_id
    cookie = cookies.pack(str(tapir_session.session_id), user_id, ip,
                          start, str(authorizations.classic))
    logger.debug('generated cookie: %s', cookie)

    user_session = domain.Session(str(tapir_session.session_id), user=user,
                                  start_time=start, end_time=end,
                                  authorizations=authorizations)
    logger.debug('created session %s', user_session.session_id)
    return user_session, cookie


def invalidate(cookie: str) -> None:
    """
    Invalidate a legacy user session.

    Parameters
    ----------
    cookie : str
        Session cookie generated when the session was created.

    Raises
    ------
    :class:`UnknownSession`
        The session could not be found, or the cookie was not valid.

    """
    try:
        session_id, user_id, ip, _, _, _ = cookies.unpack(cookie)
    except InvalidCookie as e:
        raise UnknownSession('No such session') from e

    invalidate_by_id(session_id)


def invalidate_by_id(session_id: str) -> None:
    """
    Invalidate a legacy user session by ID.

    Parameters
    ----------
    session_id : str
        Unique identifier for the session.

    Raises
    ------
    :class:`UnknownSession`
        The session could not be found, or the cookie was not valid.

    """
    delta = datetime.now(tz=EASTERN) - datetime.fromtimestamp(0, tz=EASTERN)
    end = (delta).total_seconds()
    try:
        with util.transaction() as session:
            tapir_session = _load(session_id)
            tapir_session.end_time = end - 1
            session.merge(tapir_session)
    except NoResultFound as e:
        raise UnknownSession(f'No such session {session_id}') from e
    except SQLAlchemyError as e:
        raise IOError(f'Database error') from e



"""Authn/z-related exceptions raised by components in this module."""


class InvalidToken(ValueError):
    """Token in request is not valid."""


class MissingToken(ValueError):
    """No token found in request."""


class ConfigurationError(RuntimeError):
    """The application is not configured correctly."""


class SessionCreationFailed(RuntimeError):
    """Failed to create a session in the session store."""


class SessionDeletionFailed(RuntimeError):
    """Failed to delete a session in the session store."""


class UnknownSession(RuntimeError):
    """Failed to locate a session in the session store."""


class SessionExpired(RuntimeError):
    """User's session has expired."""



"""Helpers and Flask application integration."""

from typing import Generator, Tuple, List
from datetime import datetime
from pytz import timezone
from contextlib import contextmanager
import secrets
from base64 import b64encode, b64decode
import hashlib
from collections import Counter

from sqlalchemy.engine import Engine
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

from arxiv.base.globals import get_application_config, get_application_global

from ..auth import scopes
from .. import domain
from .models import Base, DBUser, DBPolicyClass, DBEndorsement, DBSession
from .exceptions import UnknownSession, PasswordAuthenticationFailed

EASTERN = timezone('US/Eastern')


def now() -> int:
    """Get the current epoch/unix time."""
    return epoch(datetime.now(tz=EASTERN))


def epoch(t: datetime) -> int:
    """Convert a :class:`.datetime` to UNIX time."""
    delta = t - datetime.fromtimestamp(0, tz=EASTERN)
    return int(round((delta).total_seconds()))


def from_epoch(t: int) -> datetime:
    """Get a :class:`datetime` from an UNIX timestamp."""
    return datetime.fromtimestamp(t, tz=EASTERN)


@contextmanager
def transaction() -> Generator:
    """Context manager for database transaction."""
    session = current_session()
    try:
        yield session
        session.commit()
    except Exception as e:
        # logger.debug('Commit failed, rolling back: %s', str(e))
        session.rollback()
        raise


def init_app(app: object = None) -> None:
    """Set default configuration parameters for an application instance."""
    config = get_application_config(app)
    config.setdefault('CLASSIC_DATABASE_URI', 'sqlite://')


def _get_engine(app: object = None) -> Engine:
    """Get a new :class:`.Engine` for the classic database."""
    config = get_application_config(app)
    database_uri = config.get('CLASSIC_DATABASE_URI', 'sqlite://')
    return create_engine(database_uri)


def _get_session(app: object = None) -> Session:
    """Get a new :class:`.Session` for the classic database."""
    engine = _current_engine()
    return sessionmaker(bind=engine)()


def _current_engine() -> Engine:
    """Get/create :class:`.Engine` for this context."""
    g = get_application_global()
    if not g:
        return _get_engine()
    if 'user_data_engine' not in g:
        g.user_data_engine = _get_engine()
    return g.user_data_engine


def current_session() -> Session:
    """Get/create database session for this context."""
    g = get_application_global()
    if not g:
        return _get_session()
    if 'user_data' not in g:
        g.user_data = _get_session()
    return g.user_data


def create_all() -> None:
    """Create all tables in the database."""
    Base.metadata.create_all(_current_engine())


def drop_all() -> None:
    """Drop all tables in the database."""
    Base.metadata.drop_all(_current_engine())


def hash_password(password: str) -> str:
    """Generate a secure hash of a password."""
    salt = secrets.token_bytes(4)
    hashed = hashlib.sha1(salt + b'-' + password.encode('utf-8')).digest()
    return b64encode(salt + hashed).decode('ascii')


def check_password(password: str, encrypted: bytes) -> None:
    """Check a password against an encrypted hash."""
    decoded = b64decode(encrypted)
    salt = decoded[:4]
    enc_hashed = decoded[4:]
    pass_hashed = hashlib.sha1(salt + b'-' + password.encode('utf-8')).digest()
    if pass_hashed != enc_hashed:
        raise PasswordAuthenticationFailed('Incorrect password')


def compute_capabilities(tapir_user: DBUser) -> int:
    """Calculate the privilege level code for a user."""
    return int(sum([2 * tapir_user.flag_edit_users,
                    4 * tapir_user.flag_email_verified,
                    8 * tapir_user.flag_edit_system]))


def get_scopes(db_user: DBUser) -> List[str]:
    """Generate a list of authz scopes for a legacy user based on class."""
    if db_user.policy_class == DBPolicyClass.PUBLIC_USER:
        return scopes.GENERAL_USER
    return []


def is_configured() -> bool:
    """Determine whether or not the legacy database is configured."""
    config = get_application_config()
    if 'CLASSIC_DATABASE_URI' in config and 'CLASSIC_SESSION_HASH' in config:
        return True
    return False


def get_session_hash() -> str:
    """Get the legacy hash secret."""
    config = get_application_config()
    session_hash: str = config['CLASSIC_SESSION_HASH']
    return session_hash


def get_session_duration() -> int:
    """Get the session duration from the config."""
    config = get_application_config()
    timeout: str = config['CLASSIC_SESSION_TIMEOUT']
    return int(timeout)



"""
arXiv authentication and authorization tools.

This package provides core functionality for working with users and sessions
in arXiv-NG services. Housing these components in a library (separate from
service implementations) ensures that users and sessions are represented
and manipulated consistently. In addition to NG components, this package
also provides integrations with the legacy user and session data in the
classic database.

The user accounts, API client registry, and authorizer services all rely on
this package for domain representations and integration with the legacy
system.

Quick start
-----------

For typical use-cases, you will need to do the following:

1. Install this package into your virtual environment.
2. Install :class:`arxiv.users.auth.Auth` onto your application. This will make
   information about the current authenticated :class:`.domain.Session`
   available to you on the Flask request proxy object as
   ``flask.request.session``.
3. Install :class:`arxiv.users.auth.middleware.AuthMiddleware` onto your
   application. Optional (see below).

Here's an example of how you might do #2 and #3:

.. code-block:: python

   # yourapp/factory.py
   from arxiv.base import Base
   from arxiv.base.middleware import wrap
   from arxiv.users import auth


   def create_web_app() -> Flask:
       app = Flask('foo')
       Base(app)
       auth.Auth(app)    # <- Install the Auth extension.
       wrap(app, [auth.middleware.AuthMiddleware])    # <- Install middleware.
       return app

If you are not deplying this application in the cloud behind NGINX (and
therefore will not support sessions from the distributed store), you do not
need the auth middleware (step #3).
"""

from .domain import Category, UserProfile, Authorizations, UserFullName, \
    User, Session



"""
Authorization scopes for arXiv users and clients.

The concept of authorization scope comes from OAuth 2.0 (`RFC 6749 §3.3
<https://tools.ietf.org/html/rfc6749#section-3.3>`_). For a nice primer, see
`this blog post <https://brandur.org/oauth-scope>`_. The basic idea is that
the authorization associated with an access token can be limited, e.g. to
limit what actions an API client can take on behalf of a user.

In this package, the scope concept is applied to both API client and end-user
sessions. When the session is created, we consult the relevant bits of data in
our system (e.g. what roles the user has, what privileges are associated with
those roles) to determine what the user is authorized to do. Those privileges
are attached to the user's session as authorization scopes.

This module simply defines a set of constants (str) that can be used to refer
to specific authorization scopes. Rather than refer to scopes by writing new
str objects, these constants should be imported and used. This improves the
semantics of code, and reduces the risk of programming errors. For an example,
see :mod:`arxiv.users.auth.decorators`.

"""

EDIT_PROFILE = 'profile:update'
"""Authorizes editing user profile."""

VIEW_PROFILE = 'profile:read'
"""Authorizes viewing user profile."""

CREATE_SUBMISSION = 'submission:create'
"""Authorizes creating a submission."""

EDIT_SUBMISSION = 'submission:update'
VIEW_SUBMISSION = 'submission:read'
PROXY_SUBMISSION = 'submission:proxy'

GENERAL_USER = [
    EDIT_PROFILE,
    VIEW_PROFILE,
    CREATE_SUBMISSION,
    EDIT_SUBMISSION,
    VIEW_SUBMISSION
]



"""Provide an API for user authentication using the legacy database."""

from typing import Optional, Generator, Tuple
import hashlib
from base64 import b64encode, b64decode
from contextlib import contextmanager
from datetime import datetime

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound

from . import util, endorsements
from .. import domain
from ..auth import scopes
from arxiv.base import logging

from .models import Base, DBUser, DBUserPassword, DBPermanentToken, \
    DBUserNickname, DBProfile
from .exceptions import NoSuchUser, AuthenticationFailed, \
    PasswordAuthenticationFailed

logger = logging.getLogger(__name__)

PassData = Tuple[DBUser, DBUserPassword, DBUserNickname]
TokenData = Tuple[DBUser, DBPermanentToken, DBUserNickname]


def authenticate(username_or_email: Optional[str]=None,
                 password: Optional[str]=None, token: Optional[str]=None) \
        -> Tuple[domain.User, domain.Authorizations]:
    """
    Validate username/password. If successful, retrieve user details.

    Parameters
    ----------
    username_or_email : str
        Users may log in with either their username or their email address.
    password : str
        Password (as entered). Danger, Will Robinson!
    token : str
        Alternatively, the user may provide a bearer token. This is currently
        used to support "permanent" sessions, in which the token is used to
        "automatically" log the user in (i.e. without entering credentials).

    Returns
    -------
    :class:`domain.User`
    :class:`domain.Authorizations`

    Raises
    ------
    :class:`AuthenticationFailed`
        Failed to authenticate user with provided credentials.

    """
    # Users may log in using either their username or their email address.
    if username_or_email and password:
        db_user, _, db_nick = _authenticate_password(username_or_email,
                                                     password)
    # The "tapir permanent token" is effectively a bearer token. If passed,
    # a new session will be "automatically" created (from the user's
    # perspective).
    elif token:
        db_user, _, db_nick = _authenticate_token(token)
    else:
        logger.debug('Neither username/password nor token provided')
        raise AuthenticationFailed('Username+password or token required')
    user = domain.User(
        user_id=str(db_user.user_id),
        username=db_nick.nickname,
        email=db_user.email,
        name=domain.UserFullName(
            forename=db_user.first_name,
            surname=db_user.last_name,
            suffix=db_user.suffix_name
        )
    )
    auths = domain.Authorizations(
        classic=util.compute_capabilities(db_user),
        scopes=util.get_scopes(db_user),
        endorsements=endorsements.get_endorsements(user)
    )
    return user, auths


def _authenticate_token(token: str) -> TokenData:
    """
    Authenticate using a permanent token.

    Parameters
    ----------
    token : str

    Returns
    -------
    :class:`.DBUser`
    :class:`.DBPermanentToken`
    :class:`.DBUserNickname`

    Raises
    ------
    :class:`AuthenticationFailed`
        Raised if the token is malformed, or there is no corresponding token
        in the database.

    """
    try:
        user_id, secret = token.split('-')
    except ValueError as e:
        raise AuthenticationFailed('Token is malformed') from e
    try:
        return _get_token(user_id, secret)
    except NoSuchUser as e:
        logger.debug('Not a valid permanent token')
        raise AuthenticationFailed('Invalid token') from e


def _authenticate_password(username_or_email: str, password: str) -> PassData:
    """
    Authenticate using username/email and password.

    Parameters
    ----------
    username_or_email : str
        Either the email address or username of the authenticating user.
    password : str

    Returns
    -------
    :class:`.DBUser`
    :class:`.DBUserPassword`
    :class:`.DBUserNickname`

    Raises
    ------
    :class:`AuthenticationFailed`
        Raised if the user does not exist or the password is incorrect.

    """
    logger.debug(f'Authenticate with password, user: {username_or_email}')
    try:
        db_user, db_pass, db_nick = _get_user_by_username(username_or_email)
    except NoSuchUser as e:
        logger.debug(f'No such user: {username_or_email}')
        raise AuthenticationFailed('Invalid username or password') from e
    logger.debug(f'Got user with user_id: {db_user.user_id}')
    try:
        util.check_password(password, db_pass.password_enc)
    except PasswordAuthenticationFailed as e:
        raise AuthenticationFailed('Invalid username or password') from e
    return db_user, db_pass, db_nick


# TODO: look at if/how we can optimize these queries.
def _get_user_by_username(username_or_email: str) -> PassData:
    """
    Retrieve user data by username or email address.

    Parameters
    ----------
    username_or_email : str

    Returns
    -------
    :class:`.DBUser`
    :class:`.DBUserPassword`
    :class:`.DBUserNickname`

    Raises
    ------
    :class:`NoSuchUser`
        Raised when the user cannot be found.

    """
    with util.transaction() as session:
        tapir_user: DBUser = session.query(DBUser) \
            .filter(DBUser.email == username_or_email) \
            .filter(DBUser.flag_approved == 1) \
            .filter(DBUser.flag_deleted == 0) \
            .filter(DBUser.flag_banned == 0) \
            .first()
        if tapir_user:
            tapir_nick: DBUserNickname = session.query(DBUserNickname) \
                .filter(DBUserNickname.user_id == tapir_user.user_id) \
                .filter(DBUserNickname.flag_primary == 1) \
                .first()
        else:   # Usernames are stored in a separate table (!).
            tapir_nick = session.query(DBUserNickname) \
                .filter(DBUserNickname.nickname == username_or_email) \
                .filter(DBUserNickname.flag_valid == 1) \
                .first()
            if tapir_nick:
                tapir_user = session.query(DBUser) \
                    .filter(DBUser.user_id == tapir_nick.user_id) \
                    .filter(DBUser.flag_approved == 1) \
                    .filter(DBUser.flag_deleted == 0) \
                    .filter(DBUser.flag_banned == 0) \
                    .first()
        if not tapir_user:
            raise NoSuchUser('User does not exist')
        tapir_password: DBUserPassword = session.query(DBUserPassword) \
            .filter(DBUserPassword.user_id == tapir_user.user_id) \
            .first()
        if not tapir_password:
            raise RuntimeError(f'Missing password for {username_or_email}')
    return tapir_user, tapir_password, tapir_nick


def _invalidate_token(user_id: str, secret: str) -> None:
    """
    Invalidate a user's permanent login token.

    Parameters
    ----------
    user_id : str
    secret : str

    Raises
    ------
    :class:`NoSuchUser`
        Raised when the token or user cannot be found.

    """
    with util.transaction() as session:
        _, db_token, _ = _get_token(user_id, secret)
        db_token.valid = 0
        session.add(db_token)
        session.commit()


def _get_token(user_id: str, secret: str, valid: int = 1) -> TokenData:
    """
    Retrieve a user's permanent token.

    User ID and token are used together as the primary key for the token.

    Parameters
    ----------
    user_id : str
    secret : str
    valid : int
        (default: 1)

    Returns
    -------
    :class:`.DBUser`
    :class:`.DBPermanentToken`
    :class:`.DBUserNickname`

    Raises
    ------
    :class:`NoSuchUser`
        Raised when the token or user cannot be found.

    """
    with util.transaction() as session:
        db_token: DBPermanentToken = session.query(DBPermanentToken) \
            .filter(DBPermanentToken.user_id == user_id) \
            .filter(DBPermanentToken.secret == secret) \
            .filter(DBPermanentToken.valid == valid) \
            .first()    # The token must still be valid.
        if not db_token:
            raise NoSuchUser('No such token')
        if db_token:
            db_user: DBUser = session.query(DBUser) \
                .filter(DBUser.user_id == user_id) \
                .first()
        if not db_user:
            raise NoSuchUser('No user with passed token exists')
        db_nick: DBUserNickname = session.query(DBUserNickname) \
            .filter(DBUserNickname.user_id == db_user.user_id) \
            .filter(DBUserNickname.flag_primary == 1) \
            .first()
    return db_user, db_token, db_nick



"""
Functions for working with authn/z tokens on user/client requests.

Encrypted JSON Web Tokens can be used inside the arXiv system to securely
convey authn/z information about each request. These tokens will usually be
generated by the :mod:`authorizer` in response to an
authorization subrequest from the web server, and contain information about
the identity of the user and/or client as well as authorization information
(e.g. :mod:`arxiv.users.auth.scopes`).

It is essential that these JWTs are encrypted and decrypted precisely the same
way in all arXiv services, so we include these routines here for convenience.

"""

import jwt
from . import exceptions
from .. import domain


def encode(session: domain.Session, secret: str) -> str:
    """
    Encode session information as an encrypted JWT.

    Parameters
    ----------
    session : :class:`.domain.Session`
        User or client session data, including authorization information.
    secret : str
        A secret key used to encrypt the token. This secret is required to
        decode the token later on (e.g. in the application handling the
        request).

    Returns
    -------
    str
        An encrypted JWT.

    """
    return jwt.encode(domain.to_dict(session), secret).decode('ascii')


def decode(token: str, secret: str) -> domain.Session:
    """Decode an auth token to access session information."""
    try:
        data = dict(jwt.decode(token, secret, algorithms=['HS256']))
    except jwt.exceptions.DecodeError as e:  # type: ignore
        raise exceptions.InvalidToken('Not a valid token') from e
    session: domain.Session = domain.from_dict(domain.Session, data)
    return session



"""Provides functions for working with legacy session cookies."""

from typing import Tuple
from base64 import b64encode, b64decode
import hashlib
from datetime import datetime, timedelta

from .exceptions import InvalidCookie
from . import util


def unpack(cookie: str) -> Tuple[str, str, str, datetime, str]:
    """
    Unpack the legacy session cookie.

    Parameters
    ----------
    cookie : str
        The value of session cookie.

    Returns
    -------
    str
        The session ID associated with the cookie.
    str
        The user ID of the authenticated account.
    str
        The IP address of the client when the session was created.
    datetime
        The datetime when the session was created.
    datetime
        The datetime when the session expires.
    str
        Legacy user privilege level.

    Raises
    ------
    :class:`InvalidCookie`
        Raised if the cookie is detectably malformed or tampered with.

    """
    parts = cookie.split(':')
    payload: Tuple[str, str, str, datetime, str]

    if len(parts) < 5:
        raise InvalidCookie('Malformed cookie')

    session_id = parts[0]
    user_id = parts[1]
    ip = parts[2]
    issued_at = util.from_epoch(int(parts[3]))
    expires_at = issued_at + timedelta(seconds=util.get_session_duration())
    capabilities = parts[4]
    try:
        expected = pack(session_id, user_id, ip, issued_at, capabilities)
        assert expected == cookie
    except (TypeError, AssertionError) as e:
        raise InvalidCookie('Invalid session cookie; forged?') from e
    return session_id, user_id, ip, issued_at, expires_at, capabilities


def pack(session_id: str, user_id: str, ip: str, issued_at: datetime,
         capabilities: str) -> str:
    """
    Generate a value for the classic session cookie.

    Parameters
    ----------
    session_id : str
        The session ID associated with the cookie.
    user_id : str
        The user ID of the authenticated account.
    ip : str
        Client IP address.
    issued_at : datetime
        The UNIX time at which the session was initiated.
    capabilities : str
        This is essentially a user privilege level.

    Returns
    -------
    str
        Signed session cookie value.

    """
    session_hash = util.get_session_hash()
    value = ':'.join(map(str, [session_id, user_id, ip, util.epoch(issued_at),
                               capabilities]))
    to_sign = f'{value}-{session_hash}'.encode('utf-8')
    cookie_hash = b64encode(hashlib.sha256(to_sign).digest())
    return value + ':' + cookie_hash.decode('utf-8')



"""Defines user concepts for use in arXiv-NG services."""

from typing import Any, Optional, Type, NamedTuple, List, Callable
from datetime import datetime
import dateutil.parser
from pytz import timezone
from functools import partial
from arxiv import taxonomy

from arxiv.base import logging

logger = logging.getLogger(__name__)
EASTERN = timezone('US/Eastern')


class Category(NamedTuple):
    """Reprents a classification category."""

    archive: str
    """Archives group together related subjects in a domain."""

    subject: Optional[str] = None
    """Leaf-level classification in the arXiv taxonomy."""

    @classmethod
    def from_compound(cls, category: str) -> 'Category':
        """
        Create a :class:`.Category` from a compound classification slug.

        E.g. "astro-ph.CO" -> Category(archive="astro-ph", subject="CO")

        Parameters
        ----------
        category : str
            A dot-delimited compound category slug.

        Returns
        -------
        :class:`.Category`

        """
        parts = category.split('.')
        if len(parts) == 2:
            return cls(*parts)
        return cls(category, '')

    @property
    def compound(self) -> str:
        """Compound category name."""
        if self.subject:
            return f"{self.archive}.{self.subject}"
        return self.archive

    @property
    def display(self) -> str:
        """Display name for this category."""
        name: str = taxonomy.CATEGORIES[self.compound]['name']
        return name


class Client(NamedTuple):
    """Placeholder for API client."""

    client_id: str
    """Unique identifier for a :class:`.Client`."""


class UserProfile(NamedTuple):
    """User profile data."""

    # mypy (oddly) does not support class attributes:
    #  https://github.com/python/mypy/issues/3959
    STAFF = ('1', 'Staff')  # type: ignore
    PROFESSOR = ('2', 'Professor')  # type: ignore
    POST_DOC = ('3', 'Post doc')  # type: ignore
    GRAD_STUDENT = ('4', 'Grad student')  # type: ignore
    OTHER = ('5', 'Other')  # type: ignore
    RANKS = [STAFF, PROFESSOR, POST_DOC, GRAD_STUDENT, OTHER]  # type: ignore

    affiliation: str
    """Institutional affiliation."""

    country: str
    """Should be an ISO 3166-1 alpha-2 country code."""

    rank: int
    """Academic rank. Must be one of :attr:`UserProfile.RANKS`."""

    submission_groups: List[str]
    """
    Groups to which the user prefers to submit.

    Items should be one of :ref:`arxiv.taxonomy.GROUPS`.
    """

    default_category: Category
    """
    Default submission category.

    Should be one of :ref:`arxiv.taxonomy.CATEGORIES`.
    """

    homepage_url: str = ''
    """User's homepage or external profile URL."""

    remember_me: bool = True
    """Indicates whether the user prefers permanent session cookies."""

    @property
    def rank_display(self) -> str:
        """The display name of the user's rank."""
        _rank: str = dict(self.RANKS)[str(self.rank)]
        return _rank

    @property
    def default_archive(self) -> str:
        """The archive of the default category."""
        return self.default_category.archive

    @property
    def default_subject(self) -> Optional[str]:
        """The subject of the default category."""
        return self.default_category.subject

    @property
    def groups_display(self) -> str:
        """Display-ready representation of active groups for this profile."""
        return ", ".join([
            taxonomy.GROUPS[group]['name'] for group in self.submission_groups
        ])


class Authorizations(NamedTuple):
    """Authorization information, e.g. associated with a :class:`.Session`."""

    classic: int = 0
    """Capability code associated with a user's session."""

    endorsements: List[Category] = []
    """Categories to which the user is permitted to submit."""

    scopes: list = []
    """Authorized scopes. See :mod:`arxiv.users.auth.scopes`."""

    def __post_init__(self):
        """Make sure that endorsements are :class:`.Category` instances."""
        # Iterative coercion is hard. It's a lot easier to handle this here
        # than to implement a general-purpose coercsion.
        if self.endorsements and type(self.endorsements) is not Category:
            self.endorsements = [Category(**obj) for obj in self.endorsements]


class UserFullName(NamedTuple):
    """Represents a user's full name."""

    forename: str
    """First name or given name."""

    surname: str
    """Last name or family name."""

    suffix: str = ''
    """Any title or qualifier used as a suffix/postfix."""


class User(NamedTuple):
    """Represents an arXiv user and their authorizations."""

    username: str
    """Slug-like username."""

    email: str
    """The user's primary e-mail address."""

    user_id: Optional[str] = None
    """Unique identifier for the user. If ``None``, the user does not exist."""

    name: Optional[UserFullName] = None
    """The user's full name (if available)."""

    profile: Optional[UserProfile] = None
    """The user's account profile (if available)."""


class Session(NamedTuple):
    """Represents an authenticated session in the arXiv system."""

    session_id: str
    """Unique identifier for the session."""

    start_time: datetime
    """The ISO-8601 datetime when the session was created."""

    user: Optional[User] = None
    """The user for which the session was created."""

    client: Optional[Client] = None
    """The client for which the session was created."""

    end_time: Optional[datetime] = None
    """The ISO-8601 datetime when the session ended."""

    authorizations: Optional[Authorizations] = None
    """Authorizations for the current session."""

    ip_address: Optional[str] = None
    """The IP address of the client for which the session was created."""

    remote_host: Optional[str] = None
    """The hostname of the client for which the session was created."""

    nonce: Optional[str] = None
    """A pseudo-random nonce generated when the session was created."""

    @property
    def expired(self) -> bool:
        """Expired if the current time is later than :attr:`.end_time`."""
        return bool(self.end_time is not None
                    and datetime.now(tz=EASTERN) >= self.end_time)

    @property
    def expires(self) -> Optional[int]:
        """
        Number of seconds until the session expires.

        If the session is already expired, returns 0.
        """
        if self.end_time is None:
            return None
        return max((self.end_time - datetime.now(tz=EASTERN)).seconds, 0)

    # Consider for later:
    #  localization (preferred language)
    #  preferences for sharing contact info


# Helpers and private functions.


def to_dict(obj: tuple) -> dict:
    """
    Generate a dict representation of a NamedTuple instance.

    This just uses the built-in ``_asdict`` method on the intance, but also
    calls this on any child NamedTuple instances (recursively) so that the
    entire tree is cast to ``dict``.

    Parameters
    ----------
    obj : tuple
        A NamedTuple instance.

    Returns
    -------
    dict

    """
    if not hasattr(obj, '_asdict'):  # NamedTuple-generated classes have this.
        return {}
    data = obj._asdict()  # type: ignore
    _data = {}
    for key, value in data.items():
        if hasattr(value, '_asdict'):
            value = to_dict(value)
        elif isinstance(value, datetime):
            value = value.isoformat()
        _data[key] = value
    return _data


def from_dict(cls: type, data: dict) -> Any:
    """
    Generate a NamedTuple instance from a dict, with recursion.

    This is the inverse of :func:`to_dict`.

    It's easy to instantiate NamedTuples from dicts, but it's not so easy if
    a NamedTuple field is typed with another NamedTuple. This function
    instantiates a NamedTuple class ``cls`` with the data in dict ``data``,
    and also instantiates any referenced NamedTuple classes expected by
    ``cls``'s fields with the appropriate data in ``data``.

    Parameters
    ----------
    cls: type
        Any NamedTuple class.

    data: dict
        Data with which to instantiate ``cls`` and its children.

    Returns
    -------
    NamedTuple
        An instance of ``cls``.
    """
    _data = {}
    for field, field_type in cls._field_types.items():  # type: ignore
        if field not in data:
            continue
        value = data[field]
        target_type = _get_cast_type(field_type, value)
        if target_type:
            value = target_type(value)
        _data[field] = value
    return cls(**_data)


def _is_a_namedtuple(field_type: type) -> bool:
    """Determine whether or not a field type is a NamedTuple class."""
    return hasattr(field_type, '_fields')


def _is_typing_type(field_type: type) -> bool:
    """Determine whether a type is a typing class."""
    return hasattr(field_type, '_subs_tree')


def _is_nested_type(field_type: type) -> bool:
    """Determine whether a typing class is nested (e.g. Union[str, int])."""
    return type(field_type._subs_tree()) is tuple  # type: ignore


def _get_cast_type_for_str(field_type: type) -> Optional[Callable]:
    """
    Determine the target type for a ``str`` value.

    Returns ``None`` if a suitable target cannot be determined.
    """
    if (field_type is datetime or
        (_is_typing_type(field_type) and
         _is_nested_type(field_type) and
         datetime in field_type._subs_tree())):  # type: ignore
        return dateutil.parser.parse
    return None


def _get_cast_type_for_dict(field_type: type) -> Optional[Callable]:
    """
    Determine the NamedTuple target type for a ``dict`` value.

    Returns ``None`` if a suitable target cannot be determined.
    """
    # Is the expected type for the field a NamedTuple class?
    if _is_a_namedtuple(field_type):
        return partial(from_dict, field_type)

    # Is the expected type a nested Type? There may be a NamedTuple hiding
    # in there...
    if _is_typing_type(field_type) and _is_nested_type(field_type):

        # Look for either a NamedTuple class, or dict type.
        for s_type in field_type._subs_tree():  # type: ignore
            if s_type is dict:
                return None    # We already have one of these, nothing to do.
            if _is_a_namedtuple(s_type):
                return partial(from_dict, s_type)
    return None


# Recursive coersion on anything more complicated than these simple cases is
# pretty hard. It's easier to implement anything else (e.g. a List of
# something type-ish) on the domain class itself.
def _get_cast_type(field_type: type, value: Any) -> Optional[Callable]:
    """Get a casting callable for a field type/value."""
    if type(value) is dict:
        return _get_cast_type_for_dict(field_type)
    if type(value) is str:
        return _get_cast_type_for_str(field_type)
    return None



"""Provide methods for working with user accounts."""

from typing import Optional, Generator, Tuple, Any
import hashlib
from base64 import b64encode, b64decode
from contextlib import contextmanager
from datetime import datetime
import secrets

from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

from arxiv.base import logging
from .. import domain
from . import util, endorsements, exceptions, models
from .models import DBUser, DBUserPassword, DBPermanentToken, \
    DBUserNickname, DBProfile, DBPolicyClass


def username_exists(username: str) -> bool:
    """
    Determine whether a user with a particular username already exists.

    Parameters
    ----------
    username : str

    Returns
    -------
    bool

    """
    with util.transaction() as session:
        data = (
            session.query(DBUserNickname)
            .filter(DBUserNickname.nickname == username)
            .first()
        )
        if data:
            return True
        return False


def email_exists(email: str) -> bool:
    """
    Determine whether a user with a particular address already exists.

    Parameters
    ----------
    email : str

    Returns
    -------
    bool

    """
    with util.transaction() as session:
        data = session.query(DBUser).filter(DBUser.email == email).first()
        if data:
            return True
        return False


def register(user: domain.User, password: str, ip: str,
             remote_host: str) -> Tuple[domain.User, domain.Authorizations]:
    """
    Create a new user.

    Parameters
    ----------
    user : :class:`.domain.User`
        User data for the new account.
    password : str
        Password for the account.
    ip : str
        The IP address of the client requesting the registration.
    remote_host : str
        The remote hostname of the client requesting the registration.

    Returns
    -------
    :class:`.domain.User`
        Data about the created user.
    :class:`.domain.Authorizations`
        Privileges attached to the created user.

    """
    try:
        db_user, db_nick, db_profile = _create(user, password, ip, remote_host)
    except Exception as e:
        raise exceptions.RegistrationFailed('Could not create user') from e

    user = domain.User(
        user_id=str(db_user.user_id),
        username=db_nick.nickname,
        email=db_user.email,
        name=domain.UserFullName(
            forename=db_user.first_name,
            surname=db_user.last_name,
            suffix=db_user.suffix_name
        ),
        profile=db_profile.to_domain()
    )
    auths = domain.Authorizations(
        classic=util.compute_capabilities(db_user),
        scopes=util.get_scopes(db_user),
        endorsements=endorsements.get_endorsements(user)
    )
    return user, auths


def get_user_by_id(user_id: str) -> domain.User:
    """Load user data from the database."""
    db_user, db_nick, db_profile = _get_user_data(user_id)
    user = domain.User(
        user_id=str(db_user.user_id),
        username=db_nick.nickname,
        email=db_user.email,
        name=domain.UserFullName(
            forename=db_user.first_name,
            surname=db_user.last_name,
            suffix=db_user.suffix_name
        ),
        profile=db_profile.to_domain()
    )
    return user


def update(user: domain.User) -> Tuple[domain.User, domain.Authorizations]:
    """Update a user in the database."""
    if user.user_id is None:
        raise ValueError('User ID must be set')

    db_user, db_nick, db_profile = _get_user_data(user.user_id)
    with util.transaction() as session:
        _update_field(db_nick.nickname, user.username)
        _update_field(db_user.email, user.email)
        if user.name is not None:
            _update_field(db_user.first_name, user.name.forename)
            _update_field(db_user.last_name, user.name.surname)
            _update_field(db_user.suffix_name, user.name.suffix)
        if user.profile is not None:
            _update_field(db_profile.origanization, user.profile.affiliation)
            _update_field(db_profile.country, user.profile.country)
            _update_field(db_profile.rank, user.profile.rank)
            _update_field(db_profile.rank, user.profile.rank)
        session.add(db_nick)
        session.add(db_user)
        session.add(db_profile)

    user = domain.User(
        user_id=str(db_user.user_id),
        username=db_nick.nickname,
        email=db_user.email,
        name=domain.UserFullName(
            forename=db_user.first_name,
            surname=db_user.last_name,
            suffix=db_user.suffix_name
        ),
        profile=db_profile.to_domain()
    )
    auths = domain.Authorizations(
        classic=util.compute_capabilities(db_user),
        scopes=util.get_scopes(db_user),
        endorsements=endorsements.get_endorsements(user)
    )
    return user, auths


def _update_field(to_update: Any, update_with: Any) -> None:
    if to_update != update_with:
        to_update = update_with


def _get_user_data(user_id: str) -> Tuple[DBUser, DBUserNickname, DBProfile]:
    with util.transaction() as session:
        db_user, db_nick, db_profile = (
            session.query(DBUser, DBUserNickname, DBProfile)
            .filter(DBUser.user_id == user_id)
            .filter(DBUser.flag_approved == 1)
            .filter(DBUser.flag_deleted == 0)
            .filter(DBUser.flag_banned == 0)
            .filter(DBUserNickname.flag_primary == 1)
            .filter(DBUserNickname.flag_valid == 1)
            .filter(DBUserNickname.user_id == DBUser.user_id)
            .filter(DBProfile.user_id == DBUser.user_id)
            .first()
        )
        if not db_user:
            raise exceptions.NoSuchUser('User does not exist')
    return db_user, db_nick, db_profile


def _create(user: domain.User, password: str, ip: str, remote_host: str) \
        -> Tuple[DBUser, DBUserNickname, DBProfile]:
    if user.name is None:
        raise ValueError('User name must be set')
    if user.profile is None:
        raise ValueError('User profile must be set')

    with util.transaction() as session:
        # Main user entry.
        db_user = DBUser(
            first_name=user.name.forename,
            last_name=user.name.surname,
            suffix_name=user.name.suffix,
            email=user.email,
            policy_class=DBPolicyClass.PUBLIC_USER,
            joined_ip_num=ip,
            joined_remote_host=remote_host,
            joined_date=util.now(),
            tracking_cookie='',  # TODO: how to set?
        )
        session.add(db_user)

        # Nickname is (apparently) where we keep the username.
        db_nick = DBUserNickname(
            user=db_user,
            nickname=user.username,
            flag_valid=1,
            flag_primary=1
        )
        session.add(db_nick)

        def _has_group(group: str) -> int:
            if user.profile is None:
                return 0
            return int(group in user.profile.submission_groups)

        db_profile = DBProfile(
            user=db_user,
            country=user.profile.country,
            affiliation=user.profile.affiliation,
            url=user.profile.homepage_url,
            rank=user.profile.rank,
            archive=user.profile.default_archive,
            subject_class=user.profile.default_subject,
            original_subject_classes='',
            flag_group_physics=_has_group('grp_physics'),
            flag_group_math=_has_group('grp_math'),
            flag_group_cs=_has_group('grp_cs'),
            flag_group_q_bio=_has_group('grp_q-bio'),
            flag_group_q_fin=_has_group('grp_q-fin'),
            flag_group_stat=_has_group('grp_stat'),
            flag_group_eess=_has_group('grp_eess'),
            flag_group_econ=_has_group('grp_econ'),
        )
        session.add(db_profile)

        db_pass = DBUserPassword(
            user=db_user,
            password_enc=util.hash_password(password)
        )
        session.add(db_pass)
    return db_user, db_nick, db_profile



"""
Internal service API for the distributed session store.

Used to create, delete, and verify user and client session.
"""

import json
import time
import uuid
import random
from datetime import datetime, timedelta
import dateutil.parser
from pytz import timezone

from typing import Any, Optional, Union, Tuple

from functools import wraps
import redis
import jwt

from ... import domain
from ..exceptions import SessionCreationFailed, InvalidToken, \
    SessionDeletionFailed, UnknownSession

from arxiv.base.globals import get_application_config, get_application_global
from arxiv.base import logging

logger = logging.getLogger(__name__)
EASTERN = timezone('US/Eastern')


def _generate_nonce(length: int = 8) -> str:
    return ''.join([str(random.randint(0, 9)) for i in range(length)])


class SessionStore(object):
    """
    Manages a connection to Redis.

    In reality, the StrictRedis instance is thread safe, and connections are
    attached at the time a command is executed. This class simply provides a
    container for configuration.
    """

    def __init__(self, host: str, port: int, db: int, secret: str,
                 duration: int = 7200) -> None:
        """Open the connection to Redis."""
        self.r = redis.StrictRedis(host=host, port=port, db=db)
        self._secret = secret
        self._duration = duration

    def create(self, user: domain.User, authorizations: domain.Authorizations,
               ip_address: str, remote_host: str, tracking_cookie: str = '') \
            -> Tuple[domain.Session, str]:
        """
        Create a new session.

        Parameters
        ----------
        user : :class:`domain.User`
        authorizations : :class:`domain.Authorizations`

        Returns
        -------
        :class:`.Session`
        """
        session_id = str(uuid.uuid4())
        start_time = datetime.now(tz=EASTERN)
        end_time = start_time + timedelta(seconds=self._duration)
        session = domain.Session(
            session_id=session_id,
            user=user,
            start_time=start_time,
            end_time=end_time,
            authorizations=authorizations,
            nonce=_generate_nonce()
        )
        cookie = self._pack_cookie({
            'user_id': user.user_id,
            'session_id': session_id,
            'nonce': session.nonce,
            'expires': end_time.isoformat()
        })

        try:
            self.r.set(session_id, json.dumps(domain.to_dict(session)))
        except redis.exceptions.ConnectionError as e:
            raise SessionCreationFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionCreationFailed(f'Failed to create: {e}') from e

        return session, cookie

    def delete(self, session_id: str) -> None:
        """
        Delete a session in the key-value store.

        Parameters
        ----------
        session_id : str
        """
        try:
            self.r.delete(session_id)
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e

    def invalidate(self, cookie: str) -> None:
        """
        Invalidate a session in the key-value store.

        Parameters
        ----------
        cookie : str

        """
        try:
            session = domain.to_dict(self.load(cookie))
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except InvalidToken as e:
            logger.debug('Could not load session data: %s', e)
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e
        try:
            cookie_data = self._unpack_cookie(cookie)
        except jwt.exceptions.DecodeError as e:   # type: ignore
            raise SessionDeletionFailed('Bad session token') from e

        try:
            if session['nonce'] != cookie_data['nonce'] \
                    or session['user']['user_id'] != cookie_data['user_id']:
                raise SessionDeletionFailed('Bad session token')
            session['end_time'] = datetime.now(tz=EASTERN).isoformat()
            self.r.set(session['session_id'], json.dumps(session))
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e

    def invalidate_by_id(self, session_id: str) -> None:
        """
        Invalidate a session in the key-value store by ID.

        Parameters
        ----------
        session_id : str
        """
        try:
            session_data = domain.to_dict(self._load(session_id))
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except InvalidToken as e:
            logger.debug('Could not load session data: %s', e)
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e

        session_data['end_time'] = datetime.now(tz=EASTERN).isoformat()
        try:
            self.r.set(session_id, json.dumps(session_data))
        except redis.exceptions.ConnectionError as e:
            raise SessionDeletionFailed(f'Connection failed: {e}') from e
        except Exception as e:
            raise SessionDeletionFailed(f'Failed to delete: {e}') from e

    def load(self, cookie: str) -> domain.Session:
        """Load a session using a session cookie."""
        try:
            cookie_data = self._unpack_cookie(cookie)
            session_id = cookie_data['session_id']
            user_id = cookie_data.get('user_id')
            client_id = cookie_data.get('client_id')
            expires = dateutil.parser.parse(cookie_data['expires'])
            nonce = cookie_data['nonce']
        except (KeyError, jwt.exceptions.DecodeError) as e:    # type: ignore
            raise InvalidToken('Token payload malformed') from e

        if expires <= datetime.now(tz=EASTERN):
            raise InvalidToken('Session has expired')

        session = self._load(session_id)
        if session.expired:
            raise InvalidToken('Session has expired')
        if nonce != session.nonce:
            raise InvalidToken('Invalid token; likely a forgery')
        if session.user is None and session.client is None:
            raise InvalidToken('Neither user nor client data are present')
        if user_id and session.user is not None \
                and user_id != session.user.user_id:
            raise InvalidToken('Invalid token; likely a forgery')
        if client_id and session.client is not None \
                and client_id != session.client.client_id:
            raise InvalidToken('Invalid token; likely a forgery')
        return session

    def _load(self, session_id: str) -> domain.Session:
        """Get session data by session ID."""
        user_session: Union[str, bytes, bytearray] = self.r.get(session_id)
        if not user_session:
            logger.error(f'No such session: {session_id}')
            raise UnknownSession(f'Failed to find session {session_id}')
        try:
            data: dict = json.loads(user_session)
        except json.decoder.JSONDecodeError:
            raise InvalidToken('Invalid or corrupted session token')
        session: domain.Session = domain.from_dict(domain.Session, data)
        return session

    def _unpack_cookie(self, cookie: str) -> dict:
        secret = self._secret
        return dict(jwt.decode(cookie, secret))

    def _pack_cookie(self, cookie_data: dict) -> str:
        secret = self._secret
        return jwt.encode(cookie_data, secret).decode('ascii')


def init_app(app: object = None) -> None:
    """Set default configuration parameters for an application instance."""
    config = get_application_config(app)
    config.setdefault('REDIS_HOST', 'localhost')
    config.setdefault('REDIS_PORT', '6379')
    config.setdefault('REDIS_DATABASE', '0')
    config.setdefault('JWT_SECRET', 'foosecret')
    config.setdefault('SESSION_DURATION', '7200')


def get_redis_session(app: object = None) -> SessionStore:
    """Get a new session with the search index."""
    config = get_application_config(app)
    host = config.get('REDIS_HOST', 'localhost')
    port = int(config.get('REDIS_PORT', '6379'))
    db = int(config.get('REDIS_DATABASE', '0'))
    secret = config['JWT_SECRET']
    duration = int(config.get('SESSION_DURATION', '7200'))
    return SessionStore(host, port, db, secret, duration)


def current_session() -> SessionStore:
    """Get/create :class:`.SearchSession` for this context."""
    g = get_application_global()
    if not g:
        return get_redis_session()
    if 'redis' not in g:
        g.redis = get_redis_session()
    return g.redis      # type: ignore


@wraps(SessionStore.create)
def create(user: domain.User, authorizations: domain.Authorizations,
           ip_address: str, remote_host: str, tracking_cookie: str = '') \
        -> Tuple[domain.Session, str]:
    """
    Create a new session.

    Parameters
    ----------
    user : :class:`domain.User`
    authorizations : :class:`domain.Authorizations`

    Returns
    -------
    :class:`.Session`
    """
    return current_session().create(user, authorizations, ip_address,
                                    remote_host, tracking_cookie)


@wraps(SessionStore.load)
def load(cookie: str) -> domain.Session:
    """
    Load a session by cookie value.

    Parameters
    ----------
    cookie : str

    Returns
    -------
    dict
    """
    return current_session().load(cookie)


@wraps(SessionStore.delete)
def delete(session_id: str) -> None:
    """
    Delete a session in the key-value store.

    Parameters
    ----------
    session_id : str
    """
    return current_session().delete(session_id)


@wraps(SessionStore.invalidate)
def invalidate(cookie: str) -> None:
    """
    Invalidate a session in the key-value store.

    Parameters
    ----------
    cookie : str

    """
    return current_session().invalidate(cookie)


@wraps(SessionStore.invalidate_by_id)
def invalidate_by_id(session_id: str) -> None:
    """
    Invalidate a session in the key-value store by identifier.

    Parameters
    ----------
    session_id : str

    """
    return current_session().invalidate_by_id(session_id)



"""
Integration with the distributed session store.

In this implementation, we use a key-value store to hold session data
in JSON format. When a session is created, a cookie value is created (a
JSON web token) that contains information sufficient to retrieve the session.

See :mod:`.store`.
"""

from . import store



"""
Middleware for interpreting authn/z information on requestsself.

This module provides :class:`AuthMiddleware`, which unpacks encrypted JSON
Web Tokens provided via the ``Authorization`` header. This is intended to
support requests that have been pre-authorized by the web server using the
authorizer service (see :mod:`authorizer`).

The configuration parameter ``JWT_SECRET`` must be set in the WSGI request
environ (e.g. Apache's SetEnv) or in the runtime environment. This must be
the same secret that was used by the authorizer service to mint the token.

To install the middleware, use the pattern described in
:mod:`arxiv.base.middleware`. For example:

.. code-block:: python

   from arxiv.base import Base
   from arxiv.base.middleware import wrap
   from arxiv.users import auth


   def create_web_app() -> Flask:
       app = Flask('foo')
       Base(app)
       auth.Auth(app)
       wrap(app, [auth.middleware.AuthMiddleware])
       return app


For convenience, this is intended to be used with
:mod:`arxiv.users.auth.decorators`.

"""

import os
from typing import Callable, Iterable, Tuple
import jwt

from werkzeug.exceptions import Unauthorized, InternalServerError

from arxiv.base.middleware import BaseMiddleware
from arxiv.base import logging

from . import tokens
from .exceptions import InvalidToken, ConfigurationError, MissingToken
from .. import domain

logger = logging.getLogger(__name__)

WSGIRequest = Tuple[dict, Callable]


class AuthMiddleware(BaseMiddleware):
    """
    Middleware to handle auth information on requests.

    Before the request is handled by the application, the ``Authorization``
    header is parsed for an encrypted JWT. If successfully decrypted,
    information about the user and their authorization scope is attached
    to the request.

    This can be accessed in the application via
    ``flask.request.environ['session']``.  If Authorization header was not
    included, then that value will be ``None``.

    If the JWT could not be  decrypted, the value will be an
    :class:`Unauthorized` exception instance. We cannot raise the exception
    here, because the middleware is executed outside of the Flask application.
    It's up to something running inside the application (e.g.
    :func:`arxiv.users.auth.decorators.scoped`) to raise the exception.

    """

    def before(self, environ: dict, start_response: Callable) -> WSGIRequest:
        """Decode and unpack the auth token on the request."""
        environ['session'] = None      # Create the session key, at a minimum.
        environ['token'] = None
        token = environ.get('HTTP_AUTHORIZATION')    # We may not have a token.
        if token is None:
            logger.info('No auth token')
            return environ, start_response

        # The token secret should be set in the WSGI environ, or in os.environ.
        secret = environ.get('JWT_SECRET', os.environ.get('JWT_SECRET'))
        if secret is None:
            raise ConfigurationError('Missing decryption token')

        try:
            # Try to verify the token in the Authorization header, and attach
            # the decoded session data to the request.
            session: domain.Session = tokens.decode(token, secret)
            environ['session'] = session

            # Attach the encrypted token so that we can use it in subrequests.
            environ['token'] = token
        except InvalidToken as e:   # Let the application decide what to do.
            logger.error('Auth token not valid')
            exception = Unauthorized('Invalid auth token')  # type: ignore
            environ['session'] = exception
        except Exception as e:
            logger.error(f'Unhandled exception: {e}')
            exception = InternalServerError(f'Unhandled: {e}')  # type: ignore
            environ['session'] = exception
        return environ, start_response



"""Provides tools for working with authenticated user/client sessions."""

from typing import Optional, Union
from flask import Flask, request
from . import decorators, middleware, scopes, tokens
from .. import domain, legacy

from arxiv.base import logging

logger = logging.getLogger(__name__)


class Auth(object):
    """
    Attaches session and authn/z information to the request.

    Intended for use in a Flask application factory, for example:

    .. code-block:: python

       from flask import Flask
       from arxiv.users.auth import Auth
       from someapp import routes


       def create_web_app() -> Flask:
          app = Flask('someapp')
          app.config.from_pyfile('config.py')
          Auth(app)   # Registers the base/UI blueprint.
          app.register_blueprint(routes.blueprint)    # Your blueprint.
       return app


    """

    def __init__(self, app: Optional[Flask] = None) -> None:
        """
        Initialize ``app`` with base blueprint.

        Parameters
        ----------
        app : :class:`Flask`

        """
        if app is not None:
            self.init_app(app)

    def _get_legacy_session(self) -> Optional[domain.Session]:
        """
        Attempt to load a legacy auth session.

        Returns
        -------
        :class:`domain.Session` or None

        """
        classic_cookie_key = self.app.config['CLASSIC_COOKIE_NAME']
        classic_cookie = request.cookies.get(classic_cookie_key, None)
        if classic_cookie is None:
            return None
        try:
            return legacy.sessions.load(classic_cookie)
        except legacy.exceptions.UnknownSession as e:
            logger.debug('No legacy session available')
        except legacy.exceptions.InvalidCookie as e:
            logger.debug('Invalid legacy cookie')
        return None

    def init_app(self, app: Flask) -> None:
        """
        Attach :meth:`.load_session` to the Flask app.

        Parameters
        ----------
        app : :class:`Flask`

        """
        self.app = app
        self.app.before_request(self.load_session)

    def load_session(self) -> None:
        """
        Look for an active session, and attach it to the request.

        The typical scenario will involve the
        :class:`.middleware.AuthMiddleware` unpacking a session token and
        adding it to the WSGI request environ. As a fallback, if the legacy
        database is available, this method will also attempt to load an
        active legacy session.
        """
        # Check the WSGI request environ for the ``session`` key, which
        # is where the auth middleware puts any unpacked auth information from
        # the request OR any exceptions that need to be raised withing the
        # request context.
        session: Optional[Union[domain.Session, Exception]] = \
            request.environ.get('session')

        # Middlware may have passed an exception, which needs to be raised
        # within the app/execution context to be handled correctly.
        if isinstance(session, Exception):
            logger.debug('Middleware passed an exception: %s', session)
            raise session

        # If we don't see a session, we may not be deployed behind a
        # gateway with an authorization service. If the legacy database
        # is available, we can try to use that as a fall-back.
        if not session and legacy.is_configured():
            logger.debug('No session; attempting to get legacy session')
            session = self._get_legacy_session()

        # Attach the session to the request so that other
        # components can access it easily.
        request.session = session



"""Install arXiv auth package."""

from setuptools import setup, find_packages

setup(
    name='arxiv-auth',
    version='0.0.3',
    packages=[f'arxiv.{package}' for package
              in find_packages('./arxiv', exclude=['*test*'])],
    install_requires=[
        "pycountry",
        "sqlalchemy",
        "mysqlclient",
        "python-dateutil",
        "arxiv-base",
        "pyjwt",
        "redis"
    ],
    zip_safe=False
)



"""
Provide endorsement authorizations for users.

Endorsements are authorization scopes tied to specific classificatory
categories, and are used primarily to determine whether or not a user may
submit a paper with a particular primary or secondary classification.

This module preserves the behavior of the legacy system with respect to
interpreting endorsements and evaluating potential autoendorsement. The
relevant policies can be found on the `arXiv help pages
<https://arxiv.org/help/endorsement>`_.
"""

from typing import List, Dict, Optional, Callable
from collections import Counter
from datetime import datetime

from sqlalchemy.sql.expression import literal

from . import util
from .. import domain
from arxiv import taxonomy
from .models import DBUser, DBEndorsement, DBPaperOwners, DBDocuments, \
    DBDocumentInCategory, DBCategory, DBEndorsementDomain, DBEmailWhitelist, \
    DBEmailBlacklist


GENERAL_CATEGORIES = [
    domain.Category('math', 'GM'),
    domain.Category('physics', 'gen-ph')
]

WINDOW_START = util.from_epoch(157783680)


def get_endorsements(user: domain.User) -> List[domain.Category]:
    """
    Get all endorsements (explicit and implicit) for a user.

    Parameters
    ----------
    user : :class:`.domain.User`

    Returns
    -------
    list
        Each item is a :class:`.domain.Category` for which the user is
        either explicitly or implicitly endorsed.

    """
    return list(set(explicit_endorsements(user))
                | set(implicit_endorsements(user)))


def explicit_endorsements(user: domain.User) -> List[domain.Category]:
    """
    Load endorsed categories for a user.

    These are endorsements (including auto-endorsements) that have been
    explicitly commemorated.

    Parameters
    ----------
    user : :class:`.domain.User`

    Returns
    -------
    list
        Each item is a :class:`.domain.Category` for which the user is
        explicitly endorsed.
    """
    with util.transaction() as session:
        data: List[DBEndorsement] = (
            session.query(
                DBEndorsement.archive,
                DBEndorsement.subject_class,
                DBEndorsement.point_value,
            )
            .filter(DBEndorsement.endorsee_id == user.user_id)
            .filter(DBEndorsement.flag_valid == 1)
            .all()
        )
    pooled: Counter = Counter()
    for archive, subject, points in data:
        pooled[domain.Category(archive, subject)] += points
    return [category for category, points in pooled.items() if points]


def implicit_endorsements(user: domain.User) -> List[domain.Category]:
    """
    Determine categories for which a user may be autoendorsed.

    In the classic system, this was determined upon request, when the user
    attempted to submit to a particular category. Because we are separating
    authorization concerns (which includes endorsement) from the submission
    system itself, we want to calculate possible autoendorsement categories
    ahead of time.

    New development of autoendorsement-related functionality should not happen
    here. This function and related code are intended only to preserve the
    business logic already implemented in the classic system.

    Parameters
    ----------
    :class:`.User`

    Returns
    -------
    list
        Each item is a :class:`.domain.Category` for which the user may be
        auto-endorsed.
    """
    candidates = [domain.Category.from_compound(category)
                  for category, data in taxonomy.CATEGORIES_ACTIVE.items()]
    policies = category_policies()
    invalidated = invalidated_autoendorsements(user)
    papers = domain_papers(user)
    user_is_academic = is_academic(user)
    return [
        category for category in candidates
        if category in policies
        and not _disqualifying_invalidations(category, invalidated)
        and (policies[category]['endorse_all']
             or _endorse_by_email(category, policies, user_is_academic)
             or _endorse_by_papers(category, policies, papers))
    ]


def is_academic(user: domain.User) -> bool:
    """
    Determine whether a user is academic, based on their email address.

    Uses whitelist and blacklist patterns in the database.

    Parameters
    ----------
    user : :class:`.domain.User`

    Returns
    -------
    bool
    """
    with util.transaction() as session:
        in_whitelist = (
            session.query(DBEmailWhitelist)
            .filter(literal(user.email).like(DBEmailWhitelist.pattern))
            .first()
        )
        if in_whitelist:
            return True
        in_blacklist = (
            session.query(DBEmailBlacklist)
            .filter(literal(user.email).like(DBEmailBlacklist.pattern))
            .first()
        )
        if in_blacklist:
            return False
    return True


def _disqualifying_invalidations(category: domain.Category,
                                 invalidated: List[domain.Category]) -> bool:
    """
    Evaluate whether endorsement invalidations are disqualifying.

    This enforces the policy that invalidated (revoked) auto-endorsements can
    prevent future auto-endorsement.

    Parameters
    ----------
    category : :class:`.Category`
        The category for which an auto-endorsement is being considered.
    invalidated : list
        Categories for which the user has had auto-endorsements invalidated
        (revoked).

    Returns
    -------
    bool
    """
    return bool((category in GENERAL_CATEGORIES and category in invalidated)
                or (category not in GENERAL_CATEGORIES and invalidated))


def _endorse_by_email(category: domain.Category,
                      policies: Dict[domain.Category, Dict],
                      user_is_academic: bool) -> bool:
    """
    Evaluate whether an auto-endorsement can be issued based on email address.

    This enforces the policy that some categories allow auto-endorsement for
    academic users.

    Parameters
    ----------
    category : :class:`.Category`
        The category for which an auto-endorsement is being considered.
    policies : dict
        Describes auto-endorsement policies for each category (inherited from
        their endorsement domains).
    user_is_academic : bool
        Whether or not the user has been determined to be academic.

    Returns
    -------
    bool
    """
    policy = policies.get(category)
    if policy is None or 'endorse_email' not in policy:
        return False
    return policy['endorse_email'] and user_is_academic


def _endorse_by_papers(category: domain.Category,
                       policies: Dict[domain.Category, Dict],
                       papers: Dict[str, int]) -> bool:
    """
    Evaluate whether an auto-endorsement can be issued based on prior papers.

    This enforces the policy that some categories allow auto-endorsements for
    users who have published a minimum number of papers in categories that
    share an endoresement domain.

    Parameters
    ----------
    category : :class:`.Category`
        The category for which an auto-endorsement is being considered.
    policies : dict
        Describes auto-endorsement policies for each category (inherited from
        their endorsement domains).
    papers : dict
        The number of papers that the user has published in each endorsement
        domain. Keys are str names of endorsement domains, values are int.

    Returns
    -------
    bool
    """
    N_papers = papers[policies[category]['domain']]
    min_papers = policies[category]['min_papers']
    return bool(N_papers >= min_papers)


def domain_papers(user: domain.User,
                  start_date: Optional[datetime] = None) -> Dict[str, int]:
    """
    Calculate the number of papers that a user owns in each endorsement domain.

    This includes both submitted and claimed papers.

    Parameters
    ----------
    user : :class:`.domain.User`
    start_date : :class:`.datetime` or None
        If provided, will only count papers published after this date.

    Returns
    -------
    dict
        Keys are classification domains (str), values are the number of papers
        in each respective domain (int).

    """
    with util.transaction() as session:
        query = (
            session.query(
                DBPaperOwners.document_id,
                DBDocuments.document_id,
                DBDocumentInCategory.document_id,
                DBCategory.endorsement_domain
            )
            .filter(DBPaperOwners.user_id == user.user_id)
            # Lots of joins...
            .filter(DBDocuments.document_id == DBPaperOwners.document_id)
            .filter(
                DBDocumentInCategory.document_id == DBDocuments.document_id
            )
            .filter(DBCategory.archive == DBDocumentInCategory.archive)
            .filter(
                DBCategory.subject_class == DBDocumentInCategory.subject_class
            )
        )
        if start_date:
            query = query.filter(DBDocuments.dated > util.epoch(start_date))
        data = query.all()
    return dict(Counter(domain for _, _, _, domain in data).items())


def category_policies() -> Dict[domain.Category, Dict]:
    """
    Load auto-endorsement policies for each category from the database.

    Each category belongs to an endorsement domain, which defines the
    auto-endorsement policies. We retrieve those policies from the perspective
    of the individueal category for ease of lookup.

    Returns
    -------
    dict
        Keys are :class:`.domain.Category` instances. Values are dicts with
        policiy details.

    """
    with util.transaction() as session:
        data = (
            session.query(
                DBCategory.archive,
                DBCategory.subject_class,
                DBEndorsementDomain.endorse_all,
                DBEndorsementDomain.endorse_email,
                DBEndorsementDomain.papers_to_endorse,
                DBEndorsementDomain.endorsement_domain
            )
            .filter(DBCategory.definitive == 1)
            .filter(DBCategory.active == 1)
            .filter(DBCategory.endorsement_domain ==
                    DBEndorsementDomain.endorsement_domain)
            .all()
        )
    return {
        domain.Category(archive, subject): {
            'domain': e_domain,
            'endorse_all': endorse_all == 'y',
            'endorse_email': endorse_email == 'y',
            'min_papers': min_papers
        }
        for archive, subject, endorse_all, endorse_email, min_papers, e_domain
        in data
    }


def invalidated_autoendorsements(user: domain.User) -> List[domain.Category]:
    """
    Load any invalidated (revoked) auto-endorsements for a user.

    Parameters
    ----------
    user : :class:`.domain.User`

    Returns
    -------
    list
        Items are :class:`.domain.Category` for which the user has had past
        auto-endorsements revoked.
    """
    with util.transaction() as session:
        data: List[DBEndorsement] = (
            session.query(
                DBEndorsement.archive,
                DBEndorsement.subject_class
            )
            .filter(DBEndorsement.endorsee_id == user.user_id)
            .filter(DBEndorsement.flag_valid == 0)
            .filter(DBEndorsement.endorsement_type == 'auto')
            .all()
        )
    return [domain.Category(archive, subject) for archive, subject in data]



"""
Integrations with the legacy arXiv database for users and sessions.

This package provides integrations with legacy user and sessions data in the
classic DB. These components were pulled out as a separate package because
they are required by both the accounts service and the authn/z middlware,
and maintaining them in both places would create too much duplication.
"""

from . import sessions, exceptions, authenticate, models, accounts, util, \
    endorsements
from .util import create_all, init_app, current_session, drop_all, \
    is_configured



"""Exceptions for legacy user/session integration."""


class AuthenticationFailed(RuntimeError):
    """Failed to authenticate user with provided credentials."""


class NoSuchUser(RuntimeError):
    """A reference to a non-existant user was passed."""


class PasswordAuthenticationFailed(RuntimeError):
    """An invalid username/password combination were provided."""


class SessionCreationFailed(RuntimeError):
    """Failed to create a session in the legacy database."""


class SessionDeletionFailed(RuntimeError):
    """Failed to delete a session in the legacy database."""


class UnknownSession(RuntimeError):
    """Failed to locate a session in the legacy database."""


class SessionExpired(RuntimeError):
    """A reference was made to an expired session."""


class InvalidCookie(ValueError):
    """The value of a passed legacy cookie is not valid."""


class RegistrationFailed(RuntimeError):
    """Could not create a new user."""



"""
Scope-based authorization of user/client requests.

This module provides :func:`scoped`, a decorator factory used to protect Flask
routes for which authorization is required. This is done by specifying a
required authorization scope (see :mod:`arxiv.users.auth.scopes`) and/or by
providing a custom authorizer function.

Using :func:`scoped` with an authorizer function allows you to define
application-specific authorization logic on a per-request basis without adding
complexity to request controllers. The call signature of the authorizer
function should be: ``(session: domain.Session, *args, **kwargs) -> bool``,
where `*args` and `**kwargs` are the positional and keyword arguments,
respectively, passed by Flask to the decorated route function (e.g. the
URL parameters).

Here's an example of how you might use this in a Flask application:

.. code-block:: python

   from arxiv.users.auth.decorators import scoped
   from arxiv.users.auth import scopes
   from arxiv.users import domain


   def is_owner(session: domain.Session, user_id: str, **kwargs) -> bool:
       '''Check whether the authenticated user matches the requested user.'''
       return session.user.user_id == user_id


   @blueprint.route('/<string:user_id>/profile', methods=['GET'])
   @scoped(scopes.EDIT_PROFILE, authorizer=user_is_owner)
   def edit_profile(user_id: str):
       '''User can update their account information.'''
       data, code, headers = profile.get_profile(user_id)
       return render_template('accounts/profile.html', **data)


When the decorated route function is called...

- If no session is available from either the middleware or the legacy database,
  an :class:`Unauthorized` exception is raised.
- If a required scope was provided, the session is checked for the presence of
  that scope.
- If an authorization function was provided, the function is called.
- Session data is added directly to the Flask request object as
  ``request.session``, for ease of access elsewhere in the application.
- Finally, if no exceptions have been raised, the route is called with the
  original parameters.

"""

from typing import Optional, Union, Callable, Any
from functools import wraps
from flask import request
from werkzeug.exceptions import BadRequest, Unauthorized, Forbidden
from arxiv import status
from arxiv.base import logging
from .. import domain

INVALID_TOKEN = {'reason': 'Invalid authorization token'}
INVALID_SCOPE = {'reason': 'Token not authorized for this action'}


logger = logging.getLogger(__name__)


def scoped(required: Optional[str] = None,
           authorizer: Optional[Callable] = None) -> Callable:
    """
    Generate a decorator to enforce authorization requirements.

    Parameters
    ----------
    required : str
        The scope required on a user or client session in order use the
        decorated route. See :mod:`arxiv.users.auth.scopes`. If not provided,
        no scope will be enforced.
    authorizer : function
        In addition, an authorizer function may be passed to provide more
        specific authorization checks. For example, this function may check
        that the requesting user is the owner of a resource. Should have the
        signature: ``(session: domain.Session, *args, **kwargs) -> bool``.
        ``*args`` and ``**kwargs`` are the parameters passed to the decorated
        function. If the authorizer returns ``False``, an :class:`.`
        exception is raised.

    Returns
    -------
    function
        A decorator that enforces the required scope and calls the (optionally)
        provided authorizer.

    """
    def protector(func: Callable) -> Callable:
        """Decorator that provides scope enforcement."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Check the authorization token before executing the method.

            Will also raise exceptions passed by the auth middleware.

            Raises
            ------
            :class:`.Unauthorized`
                Raised when session data is not available.
            :class:`.`
                Raised when the session has insufficient auth scope, or the
                provided authorizer returns ``False``.

            """
            session = request.session
            # Use of the decorator implies that an auth session ought to be
            # present. So we'll complain here if it's not.
            if not session or not (session.user or session.client):
                logger.debug('No valid session; aborting')
                raise Unauthorized('Not a valid session')  # type: ignore

            # Check the required scopes.
            if required and (session.authorizations is None
                             or required not in session.authorizations.scopes):
                logger.debug('Session is not authorized for %s', required)
                raise Forbidden('Access denied')  # type: ignore

            # Call the provided authorizer function.
            if authorizer and not authorizer(session, *args, **kwargs):
                logger.debug('Authorizer retured negative result')
                raise Forbidden('Access denied')  # type: ignore

            logger.debug('Request is authorized, proceeding')
            return func(*args, **kwargs)
        return wrapper
    return protector
